"""Tests for the Bulb API with a light strip."""

from typing import AsyncGenerator

import pytest

from pywizlight import PilotBuilder, wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def light_strip() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP20_SHRGB_01ABI", firmware_version="1.25.0"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_setting_rgbww(light_strip: wizlight) -> None:
    """Test setting rgbww."""
    await light_strip.turn_on(PilotBuilder(rgbww=(1, 2, 3, 4, 5)))
    state = await light_strip.updateState()
    assert state and state.get_rgbww() == (1, 2, 3, 4, 5)


@pytest.mark.asyncio
async def test_supported_scenes(light_strip: wizlight) -> None:
    """Test supported scenes."""
    assert await light_strip.getSupportedScenes() == [
        "Alarm",
        "Bedtime",
        "Candlelight",
        "Christmas",
        "Cozy",
        "Cool white",
        "Daylight",
        "Diwali",
        "Deep dive",
        "Fall",
        "Fireplace",
        "Forest",
        "Focus",
        "Golden white",
        "Halloween",
        "Jungle",
        "Mojito",
        "Night light",
        "Ocean",
        "Party",
        "Pulse",
        "Pastel colors",
        "Romance",
        "Relax",
        # "Snowy sky",
        "Sunset",
        "Spring",
        "Summer",
        "Steampunk",
        "True colors",
        "TV time",
        "White",
        "Wake-up",
        "Warm white",
        "Rhythm",
    ]


@pytest.mark.asyncio
async def test_model_description_light_strip(light_strip: wizlight) -> None:
    """Test fetching the model description for a light strip."""
    bulb_type = await light_strip.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=True, color_tmp=True, effect=True, brightness=True, dual_head=False
        ),
        name="ESP20_SHRGB_01ABI",
        kelvin_range=KelvinRange(max=6500, min=2700),
        bulb_type=BulbClass.RGB,
        fw_version="1.25.0",
        white_channels=2,
        white_to_color_ratio=80,
    )



"""Exceptions for the pywizlight."""


class WizLightError(Exception):
    """General WizLightError exception occurred."""

    pass


class WizLightConnectionError(WizLightError):
    """When a connection error is encountered."""

    pass


class WizLightTimeOutError(WizLightError):
    """When a connection has timed out."""

    pass


class WizLightNotKnownBulb(WizLightError):
    """The bulb type is not known to the pywizlight."""

    pass


class WizLightMethodNotFound(WizLightError):
    """The the called bulb function is not available."""

    pass



"""Discover bulbs in a network."""

import asyncio
import json
import logging
from asyncio import AbstractEventLoop, BaseTransport, DatagramTransport, Future
from typing import Any, List, Optional, Tuple, cast

from pywizlight import wizlight
from pywizlight.models import BulbRegistry, DiscoveredBulb
from pywizlight.utils import create_udp_broadcast_socket

_LOGGER = logging.getLogger(__name__)

PORT = 38899
DEFAULT_WAIT_TIME = 5.0

# Note: The IP and address we give the bulb does not matter because
# we have register set to false which is telling the bulb to remove
# the registration
REGISTER_MESSAGE = b'{"method":"registration","params":{"phoneMac":"AAAAAAAAAAAA","register":false,"phoneIp":"1.2.3.4","id":"1"}}'  # noqa: E501


class BroadcastProtocol(asyncio.DatagramProtocol):
    """Protocol that sends an UDP broadcast message for bulb discovery."""

    def __init__(
        self,
        loop: AbstractEventLoop,
        registry: BulbRegistry,
        broadcast_address: str,
        future: Future,
    ) -> None:
        """Init discovery function."""
        self.loop = loop
        self.registry = registry
        self.broadcast_address = broadcast_address
        self.transport: Optional[DatagramTransport] = None
        self.future = future

    def connection_made(self, transport: BaseTransport) -> None:
        """Init connection to socket and register broadcasts."""
        self.transport = cast(DatagramTransport, transport)
        self.broadcast_registration()

    def broadcast_registration(self) -> None:
        """Send a registration method as UDP broadcast."""
        if not self.transport:
            return
        self.transport.sendto(REGISTER_MESSAGE, (self.broadcast_address, PORT))
        self.loop.call_later(1, self.broadcast_registration)

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Receive data from broadcast."""
        _LOGGER.debug(f"Received data {data!r} from IP address {addr}")
        try:
            resp = json.loads(data.decode())
        except json.JSONDecodeError:
            _LOGGER.error("%s: Sent invalid message: %s", addr, data)
            return
        mac = resp.get("result", {}).get("mac")
        if mac:
            _LOGGER.debug("Found bulb with IP: %s and MAC: %s", addr[0], mac)
            self.registry.register(DiscoveredBulb(addr[0], mac))

    def connection_lost(self, exc: Any) -> None:
        """Return connection error."""
        _LOGGER.debug("Closing UDP discovery")
        if exc is None:
            self.future.set_result(None)
        else:
            self.future.set_exception(exc)
        self.transport = None


async def find_wizlights(
    wait_time: float = DEFAULT_WAIT_TIME, broadcast_address: str = "255.255.255.255"
) -> List[DiscoveredBulb]:
    """Start discovery and return list of IPs of the bulbs."""
    registry = BulbRegistry()
    loop = asyncio.get_event_loop()
    future = loop.create_future()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: BroadcastProtocol(loop, registry, broadcast_address, future),
        sock=create_udp_broadcast_socket(PORT),
    )
    await asyncio.sleep(wait_time)
    transport.close()
    await future
    bulbs = registry.bulbs()
    for bulb in bulbs:
        _LOGGER.info(f"Discovered bulb {bulb.ip_address} with MAC {bulb.mac_address}")
    return bulbs


async def discover_lights(
    broadcast_space: str = "255.255.255.255", wait_time: float = DEFAULT_WAIT_TIME
) -> List[wizlight]:
    """Find lights and return list with wizlight objects."""
    discovered_IPs = await find_wizlights(
        wait_time=wait_time, broadcast_address=broadcast_space
    )
    return [
        wizlight(ip=entry.ip_address, mac=entry.mac_address) for entry in discovered_IPs
    ]



"""Tests for the Scenes."""

import pytest

from pywizlight import scenes


@pytest.mark.asyncio
async def test_get_id_from_scene_name() -> None:
    """Test looking up a scene id from the name."""
    with pytest.raises(ValueError):
        scenes.get_id_from_scene_name("non_exist")
    assert scenes.get_id_from_scene_name("Alarm") == 35



"""Tests for the Utils."""

import pytest

from pywizlight import utils


@pytest.mark.asyncio
async def test_generate_mac() -> None:
    """Test generating a mac address."""
    assert len(utils.generate_mac()) == 12


@pytest.mark.asyncio
async def test_get_source_ip() -> None:
    """Test we can get the source ip for loopback."""
    assert utils.get_source_ip("127.0.0.1") == "127.0.0.1"



"""pywizlight integration."""

import asyncio
import contextlib
import json
import logging
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from pywizlight._version import __version__ as pywizlight_version
from pywizlight.bulblibrary import BulbType
from pywizlight.exceptions import (
    WizLightConnectionError,
    WizLightMethodNotFound,
    WizLightTimeOutError,
)
from pywizlight.models import DiscoveredBulb
from pywizlight.protocol import WizProtocol
from pywizlight.push_manager import PushManager
from pywizlight.rgbcw import hs2rgbcw, rgb2rgbcw
from pywizlight.scenes import SCENES, SCENES_BY_CLASS
from pywizlight.utils import hex_to_percent, percent_to_hex, to_wiz_json
from pywizlight.vec import Vector

_LOGGER = logging.getLogger(__name__)


BulbResponse = Dict[str, Any]


PORT = 38899

# Progressive backoff config
# ===============================================================
# We send datagrams at 0, 0.75, 2.25, 5.25, 8.25, 11.25
# We wait up to 11s for the last response before declaring failure
# ================================================================
TIMEOUT = 13  # How long we will wait total
MAX_SEND_DATAGRAMS = 6  # The maximum datagrams we send
FIRST_SEND_INTERVAL = 0.75  # This is the first wait time
MAX_BACKOFF = 3  # This is how far we will backoff

PUSH_KEEP_ALIVE_INTERVAL = 20
MAX_TIME_BETWEEN_PUSH = PUSH_KEEP_ALIVE_INTERVAL + TIMEOUT


NEVER_TIME = -120.0


RGB_ORDER = ["r", "g", "b"]
RGB_COLORS = set(RGB_ORDER)
RGBW_ORDER = ["r", "g", "b", "w"]
RGBW_COLORS = set(RGBW_ORDER)
RGBWW_ORDER = ["r", "g", "b", "c", "w"]
RGBWW_COLORS = set(RGBWW_ORDER)

_IGNORE_STATE_KEYS = {"mqttCd", "ts", "rssi"}
PIR_SOURCE = "pir"

WIZMOTE_BUTTON_MAP = {
    "wfa1": "on",
    "wfa2": "off",
    "wfa3": "night",
    "wfa8": "decrease_brightness",
    "wfa9": "increase_brightness",
    "wfa16": "1",
    "wfa17": "2",
    "wfa18": "3",
    "wfa19": "4",
}

ALWAYS_SEND_SRCS = set([PIR_SOURCE, *WIZMOTE_BUTTON_MAP])

HISTORY_RECIEVE = "receive"
HISTORY_SEND = "send"
HISTORY_PUSH = "push"
HISTORY_MSG_TYPES = (HISTORY_RECIEVE, HISTORY_SEND, HISTORY_PUSH)


def states_match(old: Dict[str, Any], new: Dict[str, Any]) -> bool:
    """Check if states match except for keys we do not want to callback on."""
    old_src = old.get("src")
    new_src = new.get("src")
    # Always send the update when there is a PIR change or button press.
    if new_src != old_src and (
        new_src in ALWAYS_SEND_SRCS or old_src in ALWAYS_SEND_SRCS
    ):
        return False
    for key, val in new.items():
        if key != "src" and old.get(key) != val and key not in _IGNORE_STATE_KEYS:
            return False
    return True


def _rgb_in_range_or_raise(rgb: Tuple[Union[float, int], ...]) -> None:
    if not (0 <= rgb[0] < 256):
        raise ValueError("Red is not in range between 0-255.")
    if not (0 <= rgb[1] < 256):
        raise ValueError("Green is not in range between 0-255.")
    if not (0 <= rgb[2] < 256):
        raise ValueError("Blue is not in range between 0-255.")


def _validate_speed_or_raise(speed: int) -> None:
    if not 10 <= speed <= 200:
        raise ValueError("Value must be between 10 and 200")


def _validate_ratio_or_raise(ratio: int) -> None:
    if not 0 <= ratio <= 100:
        raise ValueError("Value must be between 0 and 100")


class PilotBuilder:
    """Get information from the bulb."""

    def __init__(
        self,
        warm_white: Optional[int] = None,
        cold_white: Optional[int] = None,
        speed: Optional[int] = None,
        scene: Optional[int] = None,
        rgb: Optional[Tuple[float, float, float]] = None,
        rgbw: Optional[Tuple[int, int, int, int]] = None,
        rgbww: Optional[Tuple[int, int, int, int, int]] = None,
        hucolor: Optional[Tuple[float, float]] = None,
        brightness: Optional[int] = None,
        colortemp: Optional[int] = None,
        state: bool = True,
        ratio: Optional[int] = None,
    ) -> None:
        """Set the parameter."""
        self.pilot_params: Dict[str, Any] = {"state": state}
        if speed is not None:
            self._set_speed(speed)
        if ratio is not None:
            self._set_ratio(ratio)
        if scene is not None:
            self._set_scene(scene)
        if brightness is not None:
            self._set_brightness(brightness)
        if rgb is not None:
            self._set_rgb(rgb)
        elif rgbw is not None:
            self._set_rgbw(rgbw)
        elif rgbww is not None:
            self._set_rgbww(rgbww)
        elif colortemp is not None:
            self._set_colortemp(colortemp)
        elif hucolor is not None:
            self._set_hs_color(hucolor)
        if warm_white is not None:
            self._set_warm_white(warm_white)
        if cold_white is not None:
            self._set_cold_white(cold_white)

    def set_pilot_message(self) -> Dict:
        """Return the pilot message."""
        return {"method": "setPilot", "params": self.pilot_params}

    def set_state_message(self, state: bool) -> Dict:
        """Return the setState message. It doesn't change the current status of the light."""
        self.pilot_params["state"] = state
        return {"method": "setState", "params": self.pilot_params}

    def _set_warm_white(self, value: int) -> None:
        """Set the value of the warm white led."""
        if not 0 <= value < 256:
            raise ValueError("Value must be between 0 and 255")
        self.pilot_params["w"] = value

    def _set_cold_white(self, value: int) -> None:
        """Set the value of the cold white led."""
        if not 0 <= value < 256:
            raise ValueError("Value must be between 0 and 255")
        self.pilot_params["c"] = value

    def _set_speed(self, value: int) -> None:
        """Set the effect changing speed (10-200)."""
        # This applies only to changing effects.
        _validate_speed_or_raise(value)
        self.pilot_params["speed"] = value

    def _set_ratio(self, value: int) -> None:
        """Set the ratio between the up and down light (1-100)."""
        _validate_ratio_or_raise(value)
        self.pilot_params["ratio"] = value

    def _set_scene(self, scene_id: int) -> None:
        """Set the scene by id."""
        if scene_id not in SCENES:
            # id not in SCENES !
            raise ValueError("Scene is not available. Only 1 to 35 are supported")
        self.pilot_params["sceneId"] = scene_id

    def _set_rgbw(self, rgbw: Tuple[int, int, int, int]) -> None:
        """Set the RGBW color state of the bulb."""
        _rgb_in_range_or_raise(rgbw)
        self.pilot_params.update({key: rgbw[idx] for idx, key in enumerate(RGB_ORDER)})
        self._set_warm_white(rgbw[3])

    def _set_rgbww(self, rgbww: Tuple[int, int, int, int, int]) -> None:
        """Set the RGBWW color state of the bulb."""
        _rgb_in_range_or_raise(rgbww)
        params = self.pilot_params
        params.update({key: rgbww[idx] for idx, key in enumerate(RGB_ORDER)})
        self._set_cold_white(rgbww[3])
        self._set_warm_white(rgbww[4])

    def _set_rgb(self, values: Tuple[float, float, float]) -> None:
        """Set the RGB color state of the bulb."""
        # Setup the tuples for the RGB values
        red, green, blue = values
        _rgb_in_range_or_raise(values)
        # Get the RGB+CW values
        rgb_out, cw = rgb2rgbcw(values)
        # Extract the RGB values
        params = self.pilot_params
        params.update({key: rgb_out[idx] for idx, key in enumerate(RGB_ORDER)})
        # No CW because of full RGB color
        if cw is not None:
            # Use the existing set_warm_white function to set the CW values
            self._set_warm_white(cw)

    def _set_hs_color(self, values: Tuple[float, float]) -> None:
        """Set the HS color state of the bulb."""
        # Transform the HS values to RGB+CW values
        rgb, cw = hs2rgbcw(values)
        _rgb_in_range_or_raise(rgb)
        # Set the RGB values
        params = self.pilot_params
        params.update({key: rgb[idx] for idx, key in enumerate(RGB_ORDER)})
        if cw is not None:
            # Use the existing set_warm_white function to set the CW values
            self._set_warm_white(cw)

    def _set_brightness(self, value: int) -> None:
        """Set the value of the brightness 0-255."""
        percent = hex_to_percent(value)
        if percent > 101:
            raise ValueError("Max value can be 100% with 255.")
        # hardware limitation - values less than 10% are not supported
        self.pilot_params["dimming"] = max(10, percent)

    def _set_colortemp(self, kelvin: int) -> None:
        """Set the color temperature for the white led in the bulb."""
        # normalize the kelvin values - should be removed
        self.pilot_params["temp"] = min(10000, max(1000, kelvin))


def _extract_bool(response: BulbResponse, key: str) -> Optional[bool]:
    return bool(response[key]) if key in response else None


def _extract_str(response: BulbResponse, key: str) -> Optional[str]:
    return str(response[key]) if key in response else None


def _extract_int(response: BulbResponse, key: str) -> Optional[int]:
    return int(response[key]) if key in response else None


class PilotParser:
    """Interpret the message from the bulb."""

    def __init__(self, pilotResult: BulbResponse) -> None:
        """Init the class."""
        self.pilotResult = pilotResult

    def get_state(self) -> Optional[bool]:
        """Return the state of the bulb."""
        return _extract_bool(self.pilotResult, "state")

    def get_source(self) -> Optional[str]:
        """Return the source of the state change."""
        return _extract_str(self.pilotResult, "src")

    def get_mac(self) -> Optional[str]:
        """Return MAC from the bulb."""
        return _extract_str(self.pilotResult, "mac")

    def get_power(self) -> Optional[float]:
        """Return power data from the bulb in watts."""
        milli_watts = _extract_int(self.pilotResult, "pc")
        return None if milli_watts is None else milli_watts / 1000

    def get_warm_white(self) -> Optional[int]:
        """Get the value of the warm white led."""
        return _extract_int(self.pilotResult, "w")

    def get_white_range(self) -> Optional[List[float]]:
        """Get the value of the whiteRange property."""
        if "whiteRange" in self.pilotResult:
            return [float(x) for x in self.pilotResult["whiteRange"]]
        return None

    def get_extended_white_range(self) -> Optional[List[float]]:
        """Get the value of the extended whiteRange property."""
        if "extRange" in self.pilotResult:
            return [float(x) for x in self.pilotResult["extRange"]]
        # New after v1.22 FW - "cctRange":[2200,2700,6500,6500]
        elif "cctRange" in self.pilotResult:
            return [float(x) for x in self.pilotResult["cctRange"]]
        return None

    def get_speed(self) -> Optional[int]:
        """Get the color changing speed."""
        return _extract_int(self.pilotResult, "speed")

    def get_ratio(self) -> Optional[int]:
        """Get the ratio between the up and down light."""
        return _extract_int(self.pilotResult, "ratio")

    def get_scene_id(self) -> Optional[int]:
        if "schdPsetId" in self.pilotResult:  # rhythm
            return 1000
        return self.pilotResult.get("sceneId")

    def get_scene(self) -> Optional[str]:
        """Get the current scene name."""
        scene_id = self.get_scene_id()
        if scene_id is None:
            return None
        return SCENES.get(scene_id)

    def get_cold_white(self) -> Optional[int]:
        """Get the value of the cold white led."""
        return _extract_int(self.pilotResult, "c")

    def get_rgb(self) -> Union[Tuple[None, None, None], Vector]:
        """Get the RGB color state of the bulb and turns it on."""
        state = self.pilotResult
        if RGB_COLORS.issubset(state):
            return tuple(int(state[val]) for val in RGB_ORDER)
        # no RGB color value was set
        return None, None, None

    def get_rgbw(self) -> Optional[Tuple[int, int, int, int]]:
        """Get the RGBW color of the device."""
        state = self.pilotResult
        if not RGBW_COLORS.issubset(state):
            return None
        return cast(
            Tuple[int, int, int, int], tuple(int(state[val]) for val in RGBW_ORDER)
        )

    def get_rgbww(self) -> Optional[Tuple[int, int, int, int, int]]:
        """Get the RGBWW color of the device."""
        state = self.pilotResult
        if not RGBWW_COLORS.issubset(state):
            return None
        return cast(
            Tuple[int, int, int, int, int],
            tuple(int(state[val]) for val in RGBWW_ORDER),
        )

    def get_brightness(self) -> Optional[int]:
        """Get the value of the brightness 0-255."""
        if "dimming" in self.pilotResult:
            return percent_to_hex(self.pilotResult["dimming"])
        return None

    def get_colortemp(self) -> Optional[int]:
        """Get the color temperature from the bulb."""
        return _extract_int(self.pilotResult, "temp")


async def _send_udp_message_with_retry(
    message: str,
    transport: asyncio.DatagramTransport,
    response_future: asyncio.Future,
    ip: str,
    port: int,
) -> None:
    """Send a UDP message multiple times until we reach the maximum or a response is recieved."""
    data = message.encode("utf-8")
    send_wait = FIRST_SEND_INTERVAL
    total_waited = 0.0
    for send in range(MAX_SEND_DATAGRAMS):
        if transport is None or transport.is_closing() or response_future.done():
            return
        attempt = send + 1
        _LOGGER.debug(
            "%s: >> %s (%s/%s) backoff=%s",
            ip,
            data,
            attempt,
            MAX_SEND_DATAGRAMS,
            total_waited,
        )
        transport.sendto(data, (ip, port))
        await asyncio.sleep(send_wait)
        total_waited += send_wait
        send_wait = min(send_wait * 2, MAX_BACKOFF)


class WizHistory:
    """Create a history instance for diagnostics."""

    def __init__(self):
        """Init the diagnostics instance."""
        self._history: Dict[str, Dict] = {
            msg_type: {} for msg_type in HISTORY_MSG_TYPES
        }
        self._last_error: Optional[str] = None

    def get(self) -> Dict:
        return {**self._history, "last_error": self._last_error}

    def error(self, msg: str) -> None:
        self._last_error = msg

    def message(self, msg_type: str, decoded: Dict) -> None:
        if "method" in decoded:
            self._history[msg_type][decoded["method"]] = decoded


class wizlight:
    """Create an instance of a WiZ Light Bulb."""

    # default port for WiZ lights - 38899
    def __init__(
        self,
        ip: str,
        port: int = PORT,
        mac: Optional[str] = None,
    ) -> None:
        """Create instance with the IP address of the bulb."""
        self.ip = ip
        self.port = port
        self.state: Optional[PilotParser] = None
        self.mac = mac
        self.bulbtype: Optional[BulbType] = None
        self.modelConfig: Optional[Dict] = None
        self.whiteRange: Optional[List[float]] = None
        self.extwhiteRange: Optional[List[float]] = None
        self.transport: Optional[asyncio.DatagramTransport] = None
        self.protocol: Optional[WizProtocol] = None
        self.history = WizHistory()

        self.lock = asyncio.Lock()
        self.loop = asyncio.get_event_loop()
        self.push_callback: Optional[Callable[[PilotParser], None]] = None
        self.response_method: Optional[str] = None
        self.response_future: Optional[asyncio.Future] = None
        self.push_cancel: Optional[Callable] = None
        self.last_push: float = NEVER_TIME
        self.push_running: bool = False
        self.power_monitoring: Optional[bool] = None
        # Check connection removed as it did blocking I/O in the event loop

    @property
    def diagnostics(self) -> dict:
        """Get diagnostics for the device."""
        return {
            "state": self.state.pilotResult if self.state else None,
            "white_range": self.whiteRange,
            "extended_white_range": self.extwhiteRange,
            "bulb_type": self.bulbtype.as_dict() if self.bulbtype else None,
            "last_push": self.last_push,
            "push_running": self.push_running,
            "version": pywizlight_version,
            "history": self.history.get(),
            "push_manager": PushManager().get().diagnostics,
        }

    @property
    def status(self) -> Optional[bool]:
        """Return the status of the bulb: true = on, false = off."""
        if self.state is None:
            return None
        return self.state.get_state()

    # ------------------ Non properties -------------- #

    async def _ensure_connection(self) -> None:
        """Ensure we are connected."""
        if self.transport:
            return
        transport_proto = await self.loop.create_datagram_endpoint(
            lambda: WizProtocol(on_response=self._on_response, on_error=self._on_error),
            remote_addr=(self.ip, self.port),
        )
        self.transport = cast(asyncio.DatagramTransport, transport_proto[0])
        self.protocol = cast(WizProtocol, transport_proto[1])

        ip, port = self.transport.get_extra_info("peername")
        if (self.ip, self.port) != (ip, port):
            _LOGGER.debug("Resolved %s:%s to %s:%s", self.ip, self.port, ip, port)
            self.ip, self.port = ip, port

    def register(self) -> None:
        """Call register to keep alive push updates."""
        if self.push_running:
            push_manager = PushManager().get()
            asyncio.create_task(self._async_send_register(push_manager.register_msg))
            self.loop.call_later(PUSH_KEEP_ALIVE_INTERVAL, self.register)

    async def _async_send_register(self, message: dict) -> None:
        """Send the registration message."""
        try:
            await self.send(message)
        except (WizLightTimeOutError, WizLightConnectionError) as ex:
            _LOGGER.debug("%s: Registration for push updates failed: %s", self.ip, ex)

    async def start_push(
        self, callback: Optional[Callable[[PilotParser], None]]
    ) -> bool:
        """Start periodic register calls to get push updates via syncPilot."""
        _LOGGER.debug("Enabling push updates for %s", self.mac)
        self.push_callback = callback
        push_manager = PushManager().get()
        self.push_cancel = push_manager.register(self.mac, self._on_push)
        if not await push_manager.start(self.ip):
            return False
        self.push_running = True
        self.register()
        return True

    def set_discovery_callback(
        self, callback: Optional[Callable[[DiscoveredBulb], None]]
    ) -> None:
        """Set the callback for when a new device is discovered."""
        PushManager().get().set_discovery_callback(callback)

    def _on_push(self, resp: dict, addr: Tuple[str, int]) -> None:
        """Handle a syncPilot from the device."""
        self.history.message(HISTORY_PUSH, resp)
        self.last_push = time.monotonic()
        old_state = self.state.pilotResult if self.state else None
        new_state = resp["params"]
        if old_state and states_match(old_state, new_state):
            return
        self.state = PilotParser(new_state)
        if self.push_callback:
            self.push_callback(self.state)

    def _on_response(self, message: bytes, addr: Tuple[str, int]) -> None:
        """Handle a response from the device."""
        _LOGGER.debug("%s: << %s", self.ip, message)
        if not self.response_future or self.response_future.done():
            return
        decoded_message = message.decode()
        try:
            decoded = json.loads(decoded_message)
        except json.JSONDecodeError as ex:
            _LOGGER.warning("%s: Failed to decode message: %s", self.ip, message)
            self.response_future.set_exception(
                WizLightConnectionError(
                    f"{self.ip}: Failed to decode json: {decoded_message}: {ex}"
                )
            )
            return
        if decoded.get("method") != self.response_method:
            # This may be a late arriving answer to a request
            # we retried since the device did not respond the first time
            _LOGGER.debug("%s: Ignoring unsolicted response: %s", self.ip, decoded)
            return
        self.response_future.set_result(decoded)

    def _on_error(self, exception: Optional[Exception]) -> None:
        """Handle a protocol error."""
        self.history.error(str(exception))
        if exception and self.response_future and not self.response_future.done():
            self.response_future.set_exception(exception)

    async def get_bulbtype(self) -> BulbType:
        """Return the bulb type as BulbType object."""
        if self.bulbtype is not None:
            return self.bulbtype

        bulb_config = await self.getBulbConfig()
        result = bulb_config["result"]
        white_range = await self.getExtendedWhiteRange()
        white_to_color_ratio = None
        white_channels = None
        if "drvConf" in result:
            # For old FW < 1.22
            white_to_color_ratio, white_channels = result["drvConf"]
        module_name = result.get("moduleName")  # Not present in 1.8.0 firmware!
        fw_version = result.get("fwVersion")
        type_id = result.get("typeId")
        model_result = self.modelConfig["result"] if self.modelConfig else {}
        white_channels = model_result.get("nowc", white_channels)
        white_to_color_ratio = model_result.get("wcr", white_to_color_ratio)
        self.bulbtype = BulbType.from_data(
            module_name,
            white_range,
            fw_version,
            white_channels,
            white_to_color_ratio,
            type_id,
        )
        return self.bulbtype

    async def getWhiteRange(self) -> Optional[List[float]]:
        """Read the white range from the bulb."""
        if self.whiteRange is not None:
            return self.whiteRange

        resp = await self.getUserConfig()
        if resp is not None and "result" in resp:
            self.whiteRange = PilotParser(resp["result"]).get_white_range()

        return self.whiteRange

    async def getExtendedWhiteRange(self) -> Optional[List[float]]:
        """Read extended withe range from the RGB bulb."""
        if self.extwhiteRange is not None:
            return self.extwhiteRange

        # First for FW > 1.22
        resp = await self.getModelConfig()
        if resp is None or "result" not in resp:
            # For old FW < 1.22
            resp = await self.getUserConfig()

        if "result" in resp:
            self.extwhiteRange = PilotParser(resp["result"]).get_extended_white_range()

        return self.extwhiteRange

    async def getSupportedScenes(self) -> List[str]:
        """Return the supported scenes based on type.

        Lookup: https://docs.pro.wizconnected.com
        """
        if self.bulbtype is None:
            await self.get_bulbtype()
        assert self.bulbtype  # Should have gotten set by get_bulbtype
        return SCENES_BY_CLASS.get(self.bulbtype.bulb_type, [])

    async def turn_off(self) -> None:
        """Turn the light off."""
        await self.send({"method": "setPilot", "params": {"state": False}})

    async def reboot(self) -> None:
        """Reboot the bulb."""
        await self.send({"method": "reboot", "params": {}})

    async def reset(self) -> None:
        """Reset the bulb to factory defaults."""
        await self.send({"method": "reset", "params": {}})

    async def set_speed(self, speed: int) -> None:
        """Set the effect speed."""
        # If we have state: True in the setPilot, the speed does not change
        _validate_speed_or_raise(speed)
        await self.send({"method": "setPilot", "params": {"speed": speed}})

    async def set_ratio(self, ratio: int) -> None:
        """Set the ratio between the up and down light."""
        _validate_ratio_or_raise(ratio)
        await self.send({"method": "setPilot", "params": {"ratio": ratio}})

    async def turn_on(self, pilot_builder: PilotBuilder = PilotBuilder()) -> None:
        """Turn the light on with defined message.

        :param pilot_builder: PilotBuilder object to set the turn on state, defaults to PilotBuilder()
        :type pilot_builder: [type], optional
        """
        await self.send(pilot_builder.set_pilot_message())

    async def set_state(self, pilot_builder: PilotBuilder = PilotBuilder()) -> None:
        """Set the state of the bulb with defined message. Doesn't turn on the light.

        :param pilot_builder: PilotBuilder object to set the state, defaults to PilotBuilder()
        :type pilot_builder: [type], optional
        """
        # TODO: self.status could be None, in which case casting it to a bool might not be what we really want
        await self.send(pilot_builder.set_state_message(bool(self.status)))

    async def get_power(self) -> Optional[float]:
        """Get watts from the device."""
        if self.last_push + MAX_TIME_BETWEEN_PUSH < time.monotonic():
            if self.power_monitoring is not False:
                try:
                    resp = await self.send({"method": "getPower"})
                except WizLightMethodNotFound:
                    self.power_monitoring = False
                    return None
                if resp is not None and "result" in resp:
                    return resp["result"]["power"] / 1000
        return self.state.get_power() if self.state else None

    # ---------- Helper Functions ------------
    async def updateState(self) -> Optional[PilotParser]:
        """Update the bulb state.

        Note: Call this method before getting any other property.
        Also, call this method to update the current value for any property.
        getPilot - gets the current bulb state - no parameters need to be included
        {"method": "getPilot", "id": 24}
        """
        if self.last_push + MAX_TIME_BETWEEN_PUSH < time.monotonic():
            resp = await self.send({"method": "getPilot", "params": {}})
            if resp is not None and "result" in resp:
                self.state = PilotParser(resp["result"])
            else:
                self.state = None
        return self.state

    def _cache_mac_from_bulb_config(self, resp: BulbResponse) -> None:
        """Cache the mac when we fetch bulb config to avoid fetching it again."""
        if resp is not None and "result" in resp and self.mac is None:
            self.mac = PilotParser(resp["result"]).get_mac()

    async def getMac(self) -> Optional[str]:
        """Read the MAC from the bulb."""
        if not self.mac:
            await self.getBulbConfig()
        return self.mac

    async def getBulbConfig(self) -> BulbResponse:
        """Return the configuration from the bulb."""
        resp = await self.send({"method": "getSystemConfig", "params": {}})
        self._cache_mac_from_bulb_config(resp)
        return resp

    async def getModelConfig(self) -> Optional[BulbResponse]:
        """Return the new model capabilities from the bulb.
        Only available in bulb FW >1.22!
        """
        if self.modelConfig is None:
            with contextlib.suppress(WizLightMethodNotFound):
                self.modelConfig = await self.send(
                    {"method": "getModelConfig", "params": {}}
                )
        return self.modelConfig

    async def getUserConfig(self) -> BulbResponse:
        """Return the user configuration from the bulb."""
        return await self.send({"method": "getUserConfig", "params": {}})

    async def lightSwitch(self) -> None:
        """Turn the light bulb on or off like a switch."""
        # first get the status
        state = await self.updateState()
        if not state:  # Did not get state, nothing to do
            return
        if state.get_state():
            # if the light is on - switch off
            await self.turn_off()
        else:
            # if the light is off - turn on
            await self.turn_on()

    async def send(self, msg_dict: Dict) -> BulbResponse:
        """Serialize a dict to json and send it to device over UDP."""
        self.history.message(HISTORY_SEND, msg_dict)
        message = to_wiz_json(msg_dict)
        await self._ensure_connection()
        assert self.transport is not None
        async with self.lock:
            self.response_method = msg_dict["method"]
            self.response_future = asyncio.Future()
            send_task = asyncio.create_task(
                _send_udp_message_with_retry(
                    message, self.transport, self.response_future, self.ip, self.port
                )
            )
            try:
                resp = await asyncio.wait_for(
                    asyncio.shield(self.response_future), timeout=TIMEOUT
                )
            except OSError as ex:
                raise WizLightConnectionError(str(ex)) from ex
            except asyncio.TimeoutError:
                _LOGGER.debug(
                    "%s: Timed out waiting for response to %s after %s tries.",
                    self.ip,
                    message,
                    MAX_SEND_DATAGRAMS,
                )
                raise WizLightTimeOutError("The request to the bulb timed out")
            finally:
                await asyncio.sleep(0)  # Try to let the task finish without cancelation
                if not send_task.done():
                    send_task.cancel()
                    with contextlib.suppress(asyncio.CancelledError):
                        await send_task
        self.history.message(HISTORY_RECIEVE, resp)
        if "error" in resp:
            if resp["error"]["code"] == -32601:
                raise WizLightMethodNotFound("Method not found; maybe older bulb FW?")
            raise WizLightConnectionError(f'Error recieved: {resp["error"]}')
        return resp

    async def async_close(self):
        """Close the transport."""
        self._async_close()

    def _async_close(self):
        self.push_running = False
        if self.transport:
            self.transport.close()
            self.transport = None
        if self.push_cancel:
            self.push_cancel()
            self.push_cancel = None

    def __del__(self):
        """Close the connection when the object is destroyed."""
        if self.transport and not self.transport.is_closing():
            self.loop.call_soon_threadsafe(self._async_close)

    def __repr__(self):
        """Return the representation of the bulb."""
        return f'{self.__class__.__name__}(ip="{self.ip}", port={self.port}, mac="{self.mac}")'



"""Tests for the Bulb API with a socket."""

import asyncio
import logging
from typing import AsyncGenerator, Optional, cast
from unittest.mock import patch

import pytest

from pywizlight import wizlight
from pywizlight.bulb import PilotParser
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.models import DiscoveredBulb
from pywizlight.protocol import WizProtocol
from pywizlight.push_manager import PushManager
from pywizlight.tests.fake_bulb import startup_bulb
from pywizlight.utils import to_wiz_json

logging.getLogger("pywizlight").setLevel(logging.DEBUG)


@pytest.fixture()
async def socket_push() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP10_SOCKET_06", firmware_version="1.25.0"
    )
    with patch("pywizlight.push_manager.RESPOND_PORT", port):
        bulb = wizlight(ip="127.0.0.1", port=port)
        yield bulb
        await bulb.async_close()
        shutdown()


@pytest.mark.asyncio
async def test_push_update_fail_no_source_ip(socket_push: wizlight) -> None:
    """Test push updates fails when we cannot get the sourrce ip."""
    last_data = PilotParser({})
    data_event = asyncio.Event()

    def _on_push(data: PilotParser) -> None:
        nonlocal last_data
        last_data = data
        data_event.set()

    with patch("pywizlight.push_manager.get_source_ip", return_value=None):
        assert await socket_push.start_push(_on_push) is False


@pytest.mark.asyncio
async def test_push_update_fail_port_in_use(socket_push: wizlight) -> None:
    """Test push updates fails when the port is in use."""
    last_data = PilotParser({})
    data_event = asyncio.Event()

    def _on_push(data: PilotParser) -> None:
        nonlocal last_data
        last_data = data
        data_event.set()

    with patch("pywizlight.push_manager.create_udp_socket", side_effect=OSError):
        assert await socket_push.start_push(_on_push) is False


@pytest.mark.asyncio
async def test_push_updates(socket_push: wizlight) -> None:
    """Test push updates."""
    bulb_type = await socket_push.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=False,
            color_tmp=False,
            effect=False,
            brightness=False,
            dual_head=False,
        ),
        name="ESP10_SOCKET_06",
        kelvin_range=KelvinRange(max=2700, min=2700),
        bulb_type=BulbClass.SOCKET,
        fw_version="1.25.0",
        white_channels=2,
        white_to_color_ratio=20,
    )
    last_data = PilotParser({})
    data_event = asyncio.Event()

    def _on_push(data: PilotParser) -> None:
        nonlocal last_data
        last_data = data
        data_event.set()

    with patch("pywizlight.push_manager.LISTEN_PORT", 0):
        assert await socket_push.start_push(_on_push) is True

    push_manager = PushManager().get()
    push_port = push_manager.push_transport.get_extra_info("sockname")[1]

    push_in_transport_proto = await asyncio.get_event_loop().create_datagram_endpoint(
        lambda: WizProtocol(on_response=lambda resp, addr: None),
        remote_addr=("127.0.0.1", push_port),
    )
    push_transport = cast(asyncio.DatagramTransport, push_in_transport_proto[0])
    params = {
        "mac": "a8bb5006033d",
        "rssi": -71,
        "src": "hb",
        "mqttCd": 255,
        "ts": 1644593327,
        "pc": 660,
        "state": False,
        "sceneId": 0,
        "temp": 6500,
        "dimming": 100,
    }

    push_transport.sendto(
        to_wiz_json(
            {
                "method": "syncPilot",
                "env": "pro",
                "params": params,
            }
        ).encode(),
        ("127.0.0.1", push_port),
    )
    await asyncio.wait_for(data_event.wait(), timeout=1)
    assert last_data.pilotResult == params
    update = await socket_push.updateState()
    assert update is not None
    assert update.pilotResult == params
    assert await socket_push.get_power() == 0.660

    diagnostics = socket_push.diagnostics
    assert diagnostics["bulb_type"]["bulb_type"] == "SOCKET"
    assert diagnostics["history"]["last_error"] is None
    assert diagnostics["push_running"] is True
    assert (
        diagnostics["history"]["push"]["syncPilot"]["params"]["mac"] == "a8bb5006033d"
    )
    assert diagnostics["push_manager"]["running"] is True
    assert diagnostics["push_manager"]["fail_reason"] is None

    push_transport.close()


@pytest.mark.asyncio
async def test_discovery_by_firstbeat(
    socket_push: wizlight, caplog: pytest.LogCaptureFixture
) -> None:
    """Test discovery from first beat."""
    bulb_type = await socket_push.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=False,
            color_tmp=False,
            effect=False,
            brightness=False,
            dual_head=False,
        ),
        name="ESP10_SOCKET_06",
        kelvin_range=KelvinRange(max=2700, min=2700),
        bulb_type=BulbClass.SOCKET,
        fw_version="1.25.0",
        white_channels=2,
        white_to_color_ratio=20,
    )
    last_discovery: Optional[DiscoveredBulb] = None
    discovery_event = asyncio.Event()

    def _on_discovery(discovery: DiscoveredBulb) -> None:
        nonlocal last_discovery
        last_discovery = discovery
        discovery_event.set()

    with patch("pywizlight.push_manager.LISTEN_PORT", 0):
        assert await socket_push.start_push(lambda data: None) is True

    assert socket_push.mac is not None
    socket_push.set_discovery_callback(_on_discovery)
    push_manager = PushManager().get()
    push_port = push_manager.push_transport.get_extra_info("sockname")[1]

    push_in_transport_proto = await asyncio.get_event_loop().create_datagram_endpoint(
        lambda: WizProtocol(on_response=lambda resp, addr: None),
        remote_addr=("127.0.0.1", push_port),
    )
    push_transport = cast(asyncio.DatagramTransport, push_in_transport_proto[0])
    push_transport.sendto(
        b"test",
        ("127.0.0.1", push_port),
    )
    push_transport.sendto(
        b"GARBAGE",
        ("127.0.0.1", push_port),
    )
    push_transport.sendto(
        to_wiz_json(
            {
                "method": "firstBeat",
                "env": "pro",
                "params": {"mac": socket_push.mac},
            }
        ).encode(),
        ("127.0.0.1", push_port),
    )
    await asyncio.wait_for(discovery_event.wait(), timeout=1)
    assert last_discovery is not None
    assert last_discovery == DiscoveredBulb("127.0.0.1", socket_push.mac)
    push_transport.close()
    assert "GARBAGE" in caplog.text



import json
import logging
import random
import socket
from typing import Optional, cast

_LOGGER = logging.getLogger(__name__)


def hex_to_percent(hex_value: float) -> float:
    """Convert hex 0-255 values to percent."""
    return round((hex_value / 255) * 100)


def percent_to_hex(percent: float) -> int:
    """Convert percent values 0-100 into hex 0-255."""
    return int(round((percent / 100) * 255))


def to_wiz_json(dump_obj):
    """Convert an object to wiz json."""
    return json.dumps(dump_obj, separators=(",", ":"))


def create_udp_broadcast_socket(listen_port: int) -> socket.socket:
    """Create a udp broadcast socket used for communicating with the device."""
    return _create_udp_socket(listen_port, reuseaddr=True, broadcast=True)


def create_udp_socket(listen_port: int) -> socket.socket:
    """Create a udp socket used for communicating with the device."""
    return _create_udp_socket(listen_port, reuseaddr=False, broadcast=False)


def _create_udp_socket(
    listen_port: int, reuseaddr: bool = False, broadcast: bool = False
) -> socket.socket:
    """Create a udp socket used for communicating with the device."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    if reuseaddr:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if broadcast:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.bind(("", listen_port))
    sock.setblocking(False)
    return sock


def generate_mac():
    """Generates a fake mac address."""
    return "{}{}{}{}{}{}{}{}{}{}{}{}".format(
        *(random.SystemRandom().choice("0123456789abcdef") for _ in range(12))
    )


def get_source_ip(target_ip: str) -> Optional[str]:
    """Return the source ip that will reach target_ip."""
    test_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    test_sock.setblocking(False)  # must be non-blocking for async
    try:
        test_sock.connect((target_ip, 1))
        return cast(str, test_sock.getsockname()[0])
    except Exception:  # pylint: disable=broad-except
        _LOGGER.debug(
            "The system could not auto detect the source ip for %s on your operating system",
            target_ip,
        )
        return None
    finally:
        test_sock.close()



"""Helper class with some math.

A bunch of utility functions, just so we don't have to bring in any external dependencies.
"""

from math import cos, sin, sqrt
from operator import add, mul, sub
from typing import Tuple

# a small value, really close to zero, more than adequate for our 3 orders of magnitude
# of color resolution
EPSILON = 1.0e-5

Vector = Tuple[float, ...]
IntVector = Tuple[int, ...]


def vecDot(a: Vector, b: Vector) -> float:
    """Retrun sum."""
    return sum(map(mul, a, b))


def vecLenSq(a: Vector) -> float:
    """Return the vector's magnitude squared."""
    return vecDot(a, a)


def vecLen(a: Vector) -> float:
    """Return the vector's magnitude."""
    lenSq = vecLenSq(a)
    return sqrt(lenSq) if (lenSq > EPSILON) else 0


def vecAdd(a: Vector, b: Vector) -> Vector:
    """Sum two vectors."""
    return tuple(map(add, a, b))


def vecSub(a: Vector, b: Vector) -> Vector:
    """Subtract two vectors."""
    return tuple(map(sub, a, b))


def vecMul(vec: Vector, sca: float) -> Vector:
    """Scale a vector."""
    return tuple(c * sca for c in vec)


def vecInt(vec: Vector) -> IntVector:
    """Truncate the vector to integer precision."""
    return tuple(map(int, vec))


def vecNormalize(vec: Vector) -> Vector:
    """Normalize the vector (i.e. make its magnitude 1)."""
    vector_length = vecLen(vec)
    return vecMul(vec, 1 / vector_length) if (vector_length > EPSILON) else vec


def vecFormat(vec: Vector) -> str:
    """Retruns something."""
    return f"({str([float(f'{n:.3f}') for n in vec])[1:-1]})"


def vecFromAngle(angle: float) -> Vector:
    """Return the unit vector for a given angle (in radians)."""
    return (cos(angle), sin(angle))



"""pywizlight integration push updates."""

import asyncio
import json
import logging
from typing import Callable, Dict, Optional, Tuple, cast

from pywizlight.models import DiscoveredBulb
from pywizlight.protocol import WizProtocol
from pywizlight.utils import create_udp_socket, generate_mac, get_source_ip

_LOGGER = logging.getLogger(__name__)

RESPOND_PORT = 38899
LISTEN_PORT = 38900


class PushManager:
    _manager = None

    @classmethod
    def get(cls):
        """Get push manager instance."""
        if cls._manager is None:
            cls._manager = cls()
        return cls._manager

    def __init__(self) -> None:
        """Initialize push manager."""
        self.push_transport: Optional[asyncio.DatagramTransport] = None
        self.push_protocol: Optional[WizProtocol] = None
        self.push_running = False
        self.discovery_callback: Optional[Callable[[DiscoveredBulb], None]] = None
        self.lock = asyncio.Lock()
        self.subscriptions: Dict[str, Callable[[Dict, Tuple[str, int]], None]] = {}
        self.register_msg: Optional[Dict] = None
        self.fail_reason: Optional[str] = None

    @property
    def diagnostics(self) -> dict:
        return {"running": self.push_running, "fail_reason": self.fail_reason}

    def set_discovery_callback(
        self, callback: Optional[Callable[[DiscoveredBulb], None]]
    ) -> None:
        """Set the callback to run when there is a new discovery."""
        self.discovery_callback = callback

    async def start(self, target_ip: str) -> bool:
        """Start listening for push updates on LISTEN_PORT if we are not already listening."""
        async with self.lock:
            if self.push_running:
                return True
            source_ip = get_source_ip(target_ip)
            if not source_ip:
                self.fail_reason = "Could not determine source ip"
                _LOGGER.warning(
                    "Could not determine source ip, falling back to polling"
                )
                return False
            try:
                sock = create_udp_socket(LISTEN_PORT)
            except OSError as ex:
                self.fail_reason = f"Port {LISTEN_PORT} is in use: {ex}"
                _LOGGER.warning(
                    "Port %s is in use: %s, cannot listen for push updates, falling back to polling",
                    LISTEN_PORT,
                    ex,
                )
                return False
            self.register_msg = {
                "params": {
                    "phoneIp": source_ip,
                    "register": True,
                    "phoneMac": generate_mac(),
                },
                "method": "registration",
            }
            push_transport_proto = (
                await asyncio.get_event_loop().create_datagram_endpoint(
                    lambda: WizProtocol(on_response=self._on_push),
                    sock=sock,
                )
            )
            self.push_transport = cast(
                asyncio.DatagramTransport, push_transport_proto[0]
            )
            self.push_protocol = cast(WizProtocol, push_transport_proto[1])
            self.push_running = True
            self.fail_reason = None
            return True

    async def stop_if_no_subs(self) -> None:
        """Stop push updates if there are no subscriptions."""
        async with self.lock:
            if not self.push_running or self.subscriptions:
                return
            self.push_running = False
            self.discovery_callback = None
            if self.push_transport:
                self.push_transport.close()
                self.push_transport = None
                self.push_protocol = None

    def register(self, mac: str, callback: Callable) -> Callable:
        """Register the subscription for a given mac address."""
        self.subscriptions[mac] = callback

        def _cancel():
            del self.subscriptions[mac]
            asyncio.create_task(self.stop_if_no_subs())

        return _cancel

    def _on_push(self, message: bytes, addr: Tuple[str, int]) -> None:
        """Handle a response from the device."""
        _LOGGER.debug("%s: PUSH << %s", addr, message)
        if message == b"test":
            return  # App sends these to test connectivity
        try:
            resp = json.loads(message.decode())
        except json.JSONDecodeError:
            _LOGGER.error("%s: Sent invalid push message: %s", addr, message)
            return
        method = resp.get("method")
        mac = resp.get("params", {}).get("mac")
        # We used to send an response to `syncPilot` messages
        # but the devices keeps sending them even if we do not
        # so we no longer send them since all it effectively
        # does it generate additional network traffic.
        if method == "firstBeat" and self.discovery_callback:
            self.discovery_callback(DiscoveredBulb(addr[0], mac))
        if method == "syncPilot" and mac in self.subscriptions:
            self.subscriptions[mac](resp, addr)



from pywizlight.bulb import PilotBuilder, PilotParser, wizlight

from pywizlight import discovery
from pywizlight.scenes import SCENES
from pywizlight.bulblibrary import BulbType

__all__ = [
    "BulbType",
    "discovery",
    "PilotBuilder",
    "PilotParser",
    "SCENES",
    "wizlight",
]



"""Manages the RGBCW color."""

import logging
from math import atan2, cos, pi
from typing import Iterable, Tuple

from .vec import (
    EPSILON,
    Vector,
    vecAdd,
    vecDot,
    vecFormat,
    vecFromAngle,
    vecInt,
    vecLen,
    vecMul,
)

_LOGGER = logging.getLogger(__name__)

ANGLE = (pi * 2) / 3
BASIS = (
    vecFromAngle(0),
    vecFromAngle(ANGLE),
    vecFromAngle(ANGLE * 2),
)
# the max value we will use for c and w
CWMAX = 128
BUFFER = ""


def debug(msg: str, end: str = "\n") -> None:
    global BUFFER
    BUFFER += msg
    if end == "\n":
        _LOGGER.debug(BUFFER)
        BUFFER = ""
    else:
        BUFFER += end


def printBasis(basis: Iterable[Vector], prefix: str = "") -> None:
    debug(f"{prefix}Basis Vectors: ", end="")
    for vector in basis:
        debug(f"{vecFormat(vector)} ", end="")
    debug("")


Trapezoid = Tuple[Vector, int]


def trapezoid(hueVec: Vector, saturation: float) -> Trapezoid:
    """This function computes the linear combination of two basis vectors that define a trapezoid.
    hueVec - a normalized vector in the hue color wheel (0..1, 0..1, 0..1)
    saturation - a single value representing the length of the hue vector (0..1)
    brightness - a separate value that may be passed in, and should be used in the Pilot.
    """
    rgb: Tuple[float, ...]
    # if saturation is essentially 0, just go to the full on
    if saturation <= EPSILON:
        rgb = (0, 0, 0)
    else:
        # we want to compute the actual RGB color of the saturated point as a linear
        # combination of no more than two of the basis vectors. first we have to figure
        # out which of the basis vectors we will use
        maxAngle = cos((pi * 2 / 3) - EPSILON)
        mask = tuple(
            (1 if (vecDot(hueVec, vector) > maxAngle) else 0) for vector in BASIS
        )
        count = sum(mask)
        debug(
            f"    Max Angle: {maxAngle:0.3f}, Mask: ({mask[0]}, {mask[1]}, {mask[2]}), Count: {count}"
        )
        if count == 1:
            # easy case, it's just one color component
            rgb = mask
        else:
            # recast as a ray-line intersection using the two found basis vectors, note
            # the basis vectors are normalized by definition
            subBasis = [BASIS[i] for i, maskVal in enumerate(mask) if (maskVal == 1)]
            printBasis(subBasis, "    ")
            # define the line from the origin along the second vector, computing its
            # equation in the form Ax + C = 0, but C is always 0 for this line
            AB = (subBasis[1][1], subBasis[1][0] * -1)
            # intersect the ray from the saturation point along the first basis vector
            # with the line we just computed, these are definitely not co-linear, so there
            # should always be an intersection point, and the result should always be in
            # the range [-1 .. 1], this is the first basis coefficient
            coeff0 = vecDot(hueVec, AB) / vecDot(subBasis[0], AB)
            # compute the intersection point, and the second basis coefficient, note that
            # we compute the coefficients to always be positive, but the intersection calculation
            # needs to be in the opposite direction from the basis vector (hence the negative on
            # coeff0).
            intersection = vecAdd(vecMul(subBasis[0], -coeff0), hueVec)
            coeff = (
                coeff0,
                vecDot(intersection, subBasis[1]),
            )
            debug(
                f"   Intersection Point: {vecFormat(intersection)}, Coefficients: {vecFormat(coeff)}"
            )
            # there's a bit of a gamut problem here, as the area outside the hexagon defined by
            # the three unit basis vectors is not actually reachable. this manifests as
            # coefficients greater than 1, which will always happen unless the target color is
            # either one of the basis vectors or a bisector of two basis vectors. we scale both
            # coefficients by 1/maxCoefficient to make valid colors
            maxCoeff = max(coeff)
            coeff = (coeff[0] / maxCoeff, coeff[1] / maxCoeff)
            debug(f"    Scaled Coefficients: {vecFormat(coeff)}")
            # now rebuild the rgb vector by putting the coefficients into the correct place
            j = 0
            rgbList = []
            for i in range(3):
                if mask[i] == 1:
                    rgbList.append(min(coeff[j], 1))
                    j += 1
                else:
                    rgbList.append(0)
            rgb = tuple(rgbList)
    # we want a discontinuous behavior. if saturation >= 0.5, we want the color to remain saturated
    # and we scale the cw value down to 0 as saturation goes from 0.5 to 1. if saturation < 0.5, we
    # want to saturate cw, and scale the rgb down to (0, 0, 0) as saturation goes from 0.5 - 0
    if saturation >= 0.5:
        # rgb remains saturated
        # scale the cw value down to 0 as saturation goes from 0.5 to 1
        cw = 1 - ((saturation - 0.5) * 2)
    else:
        cw = 1
        rgb = vecMul(rgb, saturation * 2)
    # scale back to the pilot color space
    rgb = vecInt(vecMul(rgb, 255))
    out_cw = int(max(0, cw * CWMAX))
    debug(f"    RGB OUT: {rgb}, CW: {out_cw}")
    # the wiz light appears to have 5 different LEDs, r, g, b, warm_white, and cold_white
    # there appears to be a max power supplied across the 5 LEDs, which explains why all-
    # on full isn't the brightest configuration
    # warm_white appears to be 2800k, and cold_white appears to be 6200k, somewhat neutral
    # brightness is achieved by turning both of them on
    return rgb, out_cw


def rgb2rgbcw(rgb: Vector) -> Trapezoid:
    """Convert rgb to rgbcw.
    Given a rgb tuple in the range (0..255, 0..255, 0-255), convert that to a rgbcw for the wiz
    light. brightness may or may not be passed in and is passed through to the trapezoid function.
    """
    debug(f"RGB IN: {rgb}")
    # scale the vector into canonical space ([0-1])
    rgb = vecMul(rgb, 1 / 255)
    # compute the hue vector as a linear combination of the basis vectors, and extract the
    # saturation, there's probably a better pythonese way of doing this
    hueVec = vecAdd(
        vecAdd(
            vecMul(BASIS[0], rgb[0]),
            vecMul(BASIS[1], rgb[1]),
        ),
        vecMul(BASIS[2], rgb[2]),
    )
    saturation = vecLen(hueVec)
    if saturation > EPSILON:
        hueVec = vecMul(hueVec, 1 / saturation)
    return trapezoid(hueVec, saturation)


def rgbcw2hs(rgb: Vector, cw: float) -> Tuple[float, float]:
    """Convert rgb hue.
    Given a tuple that is r,g,b and cw in 0-255 range, convert that to a hue, saturation tuple in the
    range (0..360, 0..100).
    """
    # scale the rgb and cw values into canonical space (the wiz app might set cw to higher than the
    # value we use, so we have to allow for that
    rgb = vecMul(rgb, 1 / 255)
    cw = min(cw, CWMAX) / CWMAX
    # compute the hue vector as a linear combination of the basis vectors, there's probably a
    # better pythonese way of doing this
    hueVec = vecAdd(
        vecAdd(
            vecMul(BASIS[0], rgb[0]),
            vecMul(BASIS[1], rgb[1]),
        ),
        vecMul(BASIS[2], rgb[2]),
    )
    debug(f"RGB IN: {vecFormat(rgb)}, CW: {cw:.3f}, HUE VECTOR: {vecFormat(hueVec)}")
    # the discontinuous nature of the wiz bulb setting means we have two different states:
    # 1) the cw value is 1, and the hue vector is scaled (from 50% saturation to white)
    # 2) the hue vector is saturated, and cw is scaled down (from 50% saturation to full color)
    if cw == 1:
        # hue scales down to (0, 0) at saturation 0, up to unit length at 50% saturation, so we get
        # that length, normalize the vector, and scale the saturation to reflect the half range
        hueVecLength = vecLen(hueVec)
        if hueVecLength > EPSILON:
            vecMul(hueVec, 1 / hueVecLength)
        saturation = hueVecLength * 0.5
    else:
        # the hue vector is already fully saturated, and cw scales from 0 - 0.5 to add in white light
        saturation = 1 - (cw / 2)
    # we have a saturated version of the hue vector now, which we convert to a hue vector and
    # then extract the angle of the vector in radians. We add 2 Pi to the angle if it is less than
    # 0 to put the hue angle in the range from 0 to 2 Pi
    hue = atan2(hueVec[1], hueVec[0])
    while hue < 0:
        hue += pi * 2
    # scale the hue/saturation values back to their native ranges and return the tuple
    hue *= 180 / pi
    saturation *= 100
    debug(f"    HUE OUT: {hue:.5f}, SATURATION: {saturation:.3f}")
    return hue, saturation


def hs2rgbcw(hs: Tuple[float, float]) -> Trapezoid:
    """Convert hue to a canonical value.
    given a hue, saturation tuple in the range (0..360, 0..100), convert that to a rgbcw for the wiz light
    brightness may or may not be passed in and is passed through to the trapezoid function.
    """
    hueCanonical = hs[0] / 360
    while hueCanonical >= 1:
        hueCanonical -= 1
    # compute hue in a discretized space and convert to radians, then a vector
    hueRadians = hueCanonical * pi * 2
    hueVec = vecFromAngle(hueRadians)
    # convert saturation to a canonical value in a discretized space
    # we take the square root to give the user more visual control
    saturation = hs[1] / 100
    debug(
        f"HS IN: {vecFormat(hs)}, HUE: {hueRadians:.5f}, SATURATION: {saturation:.3f}"
    )
    return trapezoid(hueVec, saturation)


def convertHSfromRGBCW(rgb: Tuple[float, ...], cw: int) -> Tuple[float, float]:
    """Convert rgb hue.
    Given a tuple that is r,g,b and cw in 0-255 range, convert that to a hue, saturation tuple in the
    range (0..360, 0..100).
    """
    red, green, blue = rgb
    if 0 <= red < 256 and 0 <= green < 256 and 0 <= blue < 256 and 0 <= cw < 255:
        return rgbcw2hs(rgb, cw)
    else:
        raise ValueError("Invalid RGB or CW values.")



"""Setup for pywizlight."""

import os
import re

import setuptools

with open("README.md") as fh:
    long_description = fh.read()


def get_version() -> str:
    """Define the version number."""
    version_file = open(os.path.join("pywizlight", "_version.py"))
    version_contents = version_file.read()
    match = re.search('__version__ = "(.*?)"', version_contents)
    assert match
    return match.group(1)


setuptools.setup(
    name="pywizlight",
    version=get_version(),
    author="Stephan Traub",
    author_email="sbidy@hotmail.com",
    description="A python connector for WiZ light bulbs (e.g SLV Play)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/sbidy/pywizlight",
    packages=setuptools.find_packages(exclude=["test.py", "tests"]),
    entry_points={"console_scripts": ["wizlight = pywizlight.cli:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["click"],
    python_requires=">=3.7",
)



"""pywizlight integration protocol."""

import asyncio
import logging
from typing import Callable, Optional, Tuple

_LOGGER = logging.getLogger(__name__)


class WizProtocol(asyncio.DatagramProtocol):
    def __init__(
        self,
        on_response: Callable[[bytes, Tuple[str, int]], None],
        on_error: Optional[Callable[[Optional[Exception]], None]] = None,
    ) -> None:
        """Init the protocol."""
        self.transport = None
        self.on_response = on_response
        self.on_error = on_error

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Trigger on_response."""
        self.on_response(data, addr)

    def error_received(self, ex: Optional[Exception]) -> None:
        """Handle error."""
        _LOGGER.debug("WizProtocol error: %s", ex)
        if self.on_error is not None:
            self.on_error(ex)

    def connection_lost(self, ex: Optional[Exception]) -> None:
        """The connection is lost."""
        _LOGGER.debug("WizProtocol connection lost: %s", ex)
